"use strict";
const service_1 = require('./model/service');
const logger_1 = require('./logger');
const ts = require("typescript");
const TreeNode_1 = require("./model/TreeNode");
var fs = require('fs');
var archy = require('archy');
class DiagramBuilder {
    constructor(filePaths) {
        this.filePaths = filePaths;
    }
    build() {
        this.buildServiceList();
    }
    writeLogsToFile(filePath) {
        logger_1.Logger.log('Saving logs to file: ' + filePath);
        fs.writeFile(filePath, logger_1.Logger.logs, (error) => {
            if (error) {
                logger_1.Logger.error('Error during file save: ' + error);
            }
            else {
                logger_1.Logger.log('File saved successfully to ' + fs.realpathSync(filePath));
            }
        });
    }
    writeServiceTreeToFile(filePath) {
        logger_1.Logger.log('Saving service tree to file: ' + filePath);
        let tree = this.printServiceTree();
        fs.writeFile(filePath, tree, (error) => {
            if (error) {
                logger_1.Logger.error('Error during file save: ' + error);
            }
            else {
                logger_1.Logger.log('File saved successfully to ' + fs.realpathSync(filePath));
            }
        });
    }
    writeServiceListToFile(filePath) {
        logger_1.Logger.log('Saving JSON list to file: ' + filePath);
        let list = this.services.map(service => {
            return {
                name: service.name,
                dependencies: service.injectedServices.map(s => s.name)
            };
        });
        fs.writeFile(filePath, JSON.stringify(list), (error) => {
            if (error) {
                logger_1.Logger.error('Error during file save: ' + error);
            }
            else {
                logger_1.Logger.log('File saved successfully to ' + fs.realpathSync(filePath));
            }
        });
    }
    printServiceTree() {
        if (!this.services || this.services.length < 1) {
            logger_1.Logger.error('No service found');
            return '';
        }
        logger_1.Logger.log(this.services.length + ' services found');
        let sortServices = (services) => {
            let s = services.sort((s1, s2) => {
                let size1 = s1.getSize();
                let size2 = s2.getSize();
                if (size1 === size2) {
                    if (s1.injectionServices.length === s2.injectionServices.length) {
                        if (s1.injectedServices.length === s2.injectedServices.length) {
                            return s1.name === s2.name ? 0 : (s1.name > s2.name ? 1 : -1);
                        }
                        return s1.injectedServices.length > s2.injectedServices.length ? 1 : -1;
                    }
                    return s1.injectionServices.length > s2.injectionServices.length ? 1 : -1;
                }
                return size1 > size2 ? 1 : -1;
            });
            return s;
        };
        let sortedServices = sortServices(this.services);
        let addServiceNodes = (service, nodeList, parents, isRoot) => {
            let cyclic = parents.some(p => service.name === p.name);
            let name = cyclic ? '!' + service.name + ' ⊂≻≺⊃' : service.name;
            if (isRoot) {
                name = '>' + name;
            }
            let node = new TreeNode_1.TreeNode('=' + name);
            nodeList.push(node);
            if (!cyclic) {
                for (let i = 0, count = service.injectionServices.length; i < count; i++) {
                    let newParents = [service];
                    parents.forEach(p => newParents.push(p));
                    addServiceNodes(service.injectionServices[i], node.nodes, newParents, false);
                }
            }
        };
        let treeNodes = [];
        for (let i = 0, count = sortedServices.length; i < count; i++) {
            addServiceNodes(sortedServices[i], treeNodes, [], true);
        }
        let outputString = '';
        let archyString = '';
        treeNodes.forEach(treeNode => archyString += archy(treeNode));
        let lines = archyString.split('\n');
        lines.forEach(line => {
            let index = line.indexOf('=');
            let isRoot = false;
            let isError = false;
            let title = line.substr(index + 1);
            if (title.startsWith('>')) {
                isRoot = true;
                title = title.substr(1);
            }
            if (title.startsWith('!')) {
                isError = true;
                title = title.substr(1);
            }
            let prefix = line.substr(0, index);
            if (isRoot && outputString.length) {
                outputString += '\n';
            }
            outputString += prefix + title + '\n';
            logger_1.Logger.logTreeElement(prefix, title, !isRoot, isError);
        });
        return outputString;
    }
    buildServiceList() {
        logger_1.Logger.log('Parsing files:\n' + this.filePaths.map(p => fs.realpathSync(p)).join('\n'));
        this.services = [];
        let program = ts.createProgram(this.filePaths, {
            target: ts.ScriptTarget.ES5, module: ts.ModuleKind.CommonJS
        });
        let checker = program.getTypeChecker();
        let isNodeExported = (node) => {
            return (node.flags & ts.NodeFlags.Export) !== 0 || (node.parent && node.parent.kind === ts.SyntaxKind.SourceFile);
        };
        let processNode = (node) => {
            if (isNodeExported(node)) {
                if (node.kind === ts.SyntaxKind.ClassDeclaration) {
                    processClass(node);
                }
                else if (node.kind === ts.SyntaxKind.ModuleDeclaration) {
                    ts.forEachChild(node, processNode);
                }
            }
        };
        let processClass = (classNode) => {
            let symbol = checker.getSymbolAtLocation(classNode.name);
            let className = symbol.getName();
            if (classNode.decorators && classNode.decorators.some(decorator => {
                let symbol = checker.getSymbolAtLocation(decorator.expression.getFirstToken());
                return symbol && symbol.getName() === 'Injectable';
            })) {
                logger_1.Logger.log('Service: ' + className);
                let service = new service_1.Service(className);
                this.services.push(service);
                let constructorType = checker.getTypeOfSymbolAtLocation(symbol, symbol.valueDeclaration);
                let constructors = constructorType.getConstructSignatures();
                constructors.forEach(constructor => {
                    let params = constructor.getParameters();
                    params.forEach(p => {
                        let paramType = checker.typeToString(checker.getTypeOfSymbolAtLocation(p, p.valueDeclaration));
                        if (!service.injectedServices.some(is => is.name === paramType)) {
                            service.injectedServices.push(new service_1.Service(paramType));
                        }
                    });
                });
            }
            else {
                logger_1.Logger.log('Ignored class: ' + className);
            }
        };
        for (const sourceFile of program.getSourceFiles()) {
            ts.forEachChild(sourceFile, processNode);
        }
        this.services.forEach(service => {
            let first = true;
            let index = -1;
            while (first || index >= 0) {
                index = service.injectedServices.findIndex(is => !this.services.some(s => s.name === is.name));
                first = false;
                if (index >= 0) {
                    logger_1.Logger.log('Ignored: ' + service.injectedServices[index].name);
                    service.injectedServices.splice(index, 1);
                }
            }
        });
        this.services.forEach(service => {
            service.injectionServices = this.services.filter(s => {
                return s !== service && s.injectedServices.some(is => is.name === service.name);
            });
        });
    }
}
exports.DiagramBuilder = DiagramBuilder;
//# sourceMappingURL=builder.js.map